#define ColorStone    &hff7700
#define ColorCrystal  &h0000ff
#define ColorCreature &h00ff00
#define MAXEFFECTS    4
#define MAXCARDS      6

enum TEffectType
  none = 0
  changeOwnStone, changeOwnCrystal, changeOwnCreature, changeOwnStoneProd, changeOwnCrystalProd, changeOwnCreatureProd
  changeOppStone, changeOppCrystal, changeOppCreature, changeOppStoneProd, changeOppCrystalProd, changeOppCreatureProd
  changeBothStone, changeBothCrystal, changeBothCreature, changeBothStoneProd, changeBothCrystalProd, changeBothCreatureProd
  stealStone, stealCrystal, stealCreature, stealStoneProd, stealCrystalProd, stealCreatureProd
  changeOwnWall, changeOppWall, changeBothWall, changeOwnTower, changeOppTower, changeBothTower
  ownDamage, oppDamage, bothDamage
  damageOwnWall, damageOppWall, damageBothWall, damageOwnTower, damageOppTower, damageBothTower
end enum
enum TCond
  none = 0
  ownStone,  ownCrystal,  ownCreature,  ownStoneProd,  ownCrystalProd,  ownCreatureProd,  ownWall,  ownTower
  oppStone,  oppCrystal,  oppCreature,  oppStoneProd,  oppCrystalProd,  oppCreatureProd,  oppWall,  oppTower
  diffStone, diffCrystal, diffCreature, diffStoneProd, diffCrystalProd, diffCreatureProd, diffWall, diffTower
  diffOwnBuilding, diffOppBuilding
end enum
enum TComp
  none = 0
  equal, greater, lower
end enum
enum TCommand
  ok = 128, fault, wait4connection, tooManyClients, connectionLost, ready, restart
  submitSettings, playCard, dropCard, takeCard, yourTurn, wait4turn
  gameWin, gameLose, gameDraw, withdraw
end enum
enum TError
  none = 0
  ' error when trying to create an ai client
  serverFull, connectionError
  ' error when playing a card
  notYourTurn, cardNotPlayable
  ' misc
  missingParameter, wrongParameter, unknown
end enum
namespace Special
  dim as integer nothing = 0, again = 1, stone = 2, crystal = 4, creature = 6, ongoing = 8
end namespace

type TGameSettings
  as integer waitBeforeDisplay, waitAfterDisplay
  as integer startPStone, startPCrystal, startPCreature
  as integer startStone, startCrystal, startCreature
  as integer maxStone, maxCrystal, maxCreature
  as integer startTower, startWall, maxTower
  as string*256 fillTower, fillWall
end type
dim shared as TGameSettings setting

type TEffect
  as TEffectType effectType    ' kind of effect
  as TEffectType effectTypeAlt ' kind of alternative effect
  as integer power, powerAlt   ' power of the effect
  as TCond relation, relatAlt  ' relation
  as TCond cond1, cond2        ' kind of condition
  as TComp comp                ' kind of comparsion
  as integer compValue         ' value to compare with
  as string desc, descAlt      ' optional description
  ' if (cond1 comp [cond2|compValue]) then effectType(power) else effectTypeAlt(powerAlt) 
end type

type TCard
  as string  cardname, desc
  as ushort  id
  as integer cost, costType, countEffects
  as TEffect effect(MAXEFFECTS-1)
  as integer special
  declare sub addEffect(e as TEffectType, p as integer, _
              cond1 as TCond = TCond.none, cond2 as TCond = TCond.none, comp as TComp = TComp.none, _
              v as integer = 0, relation as TCond = TCond.none)
end type
sub TCard.addEffect(e as TEffectType, p as integer, _
          cond1 as TCond = TCond.none, cond2 as TCond = TCond.none, comp as TComp = TComp.none, _
          v as integer = 0, relation as TCond = TCond.none)
  if this.countEffects < ubound(this.effect) then
    with this.effect(this.countEffects)
      .effectType = e
      .power      = p
      .cond1      = cond1
      .cond2      = cond2
      .comp       = comp
      .compValue  = v
      .relation   = relation
    end with
    this.countEffects += 1
  end if
end sub

type TPlayer
  as integer pStone, pCrystal, pCreature                 ' production
  as integer stone, crystal, creature                    ' resources
  as integer wall, tower, card(MAXCARDS-1)
  as TCard deck(any)
  declare constructor
  declare sub production
end type
constructor TPlayer
  this.pStone    = setting.startPStone
  this.pCrystal  = setting.startPCrystal
  this.pCreature = setting.startPCreature
  this.stone     = setting.startStone
  this.crystal   = setting.startCrystal
  this.creature  = setting.startCreature
  this.tower     = setting.startTower
  this.wall      = setting.startWall
end constructor
sub TPlayer.production
  this.stone    += this.pStone
  this.crystal  += this.pCrystal
  this.creature += this.pCreature
end sub

type TDeck
  as TDeck ptr prev, nxt
  as integer card
  declare sub add(card as integer)
  declare sub del
end type
sub TDeck.add(card as integer)
  if this.nxt = 0 then
    this.prev = @this
    this.nxt  = @this
    this.card = card
  else
    dim as TDeck ptr newCard = new TDeck
    newCard->prev = @this
    newCard->nxt  = this.nxt
    newCard->card = card
    this.nxt->prev = newCard
    this.nxt = newCard
  end if
end sub
sub TDeck.del
  dim as TDeck ptr nxt = this.nxt
  if nxt = @this then
    this.prev = 0
    this.nxt  = 0
    this.card = 0
  else
    nxt->nxt->prev = @this
    this.nxt = nxt->nxt
    delete nxt
  end if
end sub

declare sub calcEffect(player as TPlayer, opponent as TPlayer, effect as TEffect)
declare function calcRelation(player as TPlayer, opponent as TPlayer, value as integer, rel as TCond) as integer
declare function getConditionValue(cond as string) as TCond
declare sub readCards(player as TPlayer, filename as string, language as string)
declare function readEffect(player as TPlayer, byref c as integer, xml as TypXML) as TEffectType
declare function removeNewline(text as string) as string
declare function testCondition(player as TPlayer, opponent as TPlayer, e as TEffect) as integer

sub calcEffect(player as TPlayer, opponent as TPlayer, effect as TEffect)
  dim p as integer, r as TCond, e as TEffectType
  with effect
    if testCondition(player, opponent, effect) then
      e = .effectType
      p =  calcRelation(player, opponent, .power, .relation)
    else
      e = .effectTypeAlt
      p =  calcRelation(player, opponent, .powerAlt, .relatAlt)
    end if
    if r <> TCond.none then p = calcRelation(player, opponent, p, r)
    select case e
      case TEffectType.stealStone
        if p > 0 then
          if opponent.stone < p then p = opponent.stone
        else
          if player.stone < -p then p = -player.stone
        end if
      case TEffectType.stealCrystal
        if p > 0 then
          if opponent.crystal < p then p = opponent.crystal
        else
          if player.crystal < -p then p = -player.crystal
        end if
      case TEffectType.stealCreature
        if p > 0 then
          if opponent.creature < p then p = opponent.creature
        else
          if player.creature < -p then p = -player.creature
        end if
      case TEffectType.stealStoneProd
        if p > 0 then
          if opponent.pStone < p+1 then p = opponent.pStone-1
        else
          if player.pStone < -p+1 then p = -player.pStone-1
        end if
      case TEffectType.stealCrystalProd
        if p > 0 then
          if opponent.pCrystal < p+1 then p = opponent.pCrystal-1
        else
          if player.pCrystal < -p+1 then p = -player.pCrystal-1
        end if
      case TEffectType.stealCreatureProd
        if p > 0 then
          if opponent.pCreature < p+1 then p = opponent.pCreature-1
        else
          if player.pCreature < -p+1 then p = -player.pCreature-1
        end if
    end select
    select case e
      case TEffectType.changeOwnStone         : player.stone       += p
      case TEffectType.changeOwnCrystal       : player.crystal     += p
      case TEffectType.changeOwnCreature      : player.creature    += p
      case TEffectType.changeOwnStoneProd     : player.pStone      += p
      case TEffectType.changeOwnCrystalProd   : player.pCrystal    += p
      case TEffectType.changeOwnCreatureProd  : player.pCreature   += p
      case TEffectType.changeOppStone         : opponent.stone     += p
      case TEffectType.changeOppCrystal       : opponent.crystal   += p
      case TEffectType.changeOppCreature      : opponent.creature  += p
      case TEffectType.changeOppStoneProd     : opponent.pStone    += p
      case TEffectType.changeOppCrystalProd   : opponent.pCrystal  += p
      case TEffectType.changeOppCreatureProd  : opponent.pCreature += p
      case TEffectType.changeBothStone        : player.stone       += p : opponent.stone     += p
      case TEffectType.changeBothCrystal      : player.crystal     += p : opponent.crystal   += p
      case TEffectType.changeBothCreature     : player.creature    += p : opponent.creature  += p
      case TEffectType.changeBothStoneProd    : player.pStone      += p : opponent.pStone    += p
      case TEffectType.changeBothCrystalProd  : player.pCrystal    += p : opponent.pCrystal  += p
      case TEffectType.changeBothCreatureProd : player.pCreature   += p : opponent.pCreature += p
      case TEffectType.stealStone             : player.stone       += p : opponent.stone     -= p
      case TEffectType.stealCrystal           : player.crystal     += p : opponent.crystal   -= p
      case TEffectType.stealCreature          : player.creature    += p : opponent.creature  -= p
      case TEffectType.stealStoneProd         : player.pStone      += p : opponent.pStone    -= p
      case TEffectType.stealCrystalProd       : player.pCrystal    += p : opponent.pCrystal  -= p
      case TEffectType.stealCreatureProd      : player.pCreature   += p : opponent.pCreature -= p
      case TEffectType.changeOwnWall          : player.wall        += p
      case TEffectType.changeOppWall          : opponent.wall      += p
      case TEffectType.changeBothWall         : player.wall        += p : opponent.wall   += p
      case TEffectType.changeOwnTower         : player.tower       += p
      case TEffectType.changeOppTower         : opponent.tower     += p
      case TEffectType.changeBothTower        : player.tower       += p : opponent.tower  += p
      case TEffectType.ownDamage
        if player.wall >= p then
          player.wall -= p
        else
          player.tower += player.wall - p
          player.wall = 0
        end if
      case TEffectType.oppDamage
        if opponent.wall >= p then
          opponent.wall -= p
        else
          opponent.tower += opponent.wall - p
          opponent.wall = 0
        end if
      case TEffectType.bothDamage
        if player.wall >= p then
          player.wall -= p
        else
          player.tower += player.wall - p
          player.wall = 0
        end if
        if opponent.wall >= p then
          opponent.wall -= p
        else
          opponent.tower += opponent.wall - p
          opponent.wall = 0
        end if
    end select
  end with
end sub

function calcRelation(player as TPlayer, opponent as TPlayer, value as integer, rel as TCond) as integer
  select case rel
    case ownStone         : return (value * player.stone)       \ 100
    case ownCrystal       : return (value * player.crystal)     \ 100
    case ownCreature      : return (value * player.creature)    \ 100
    case ownStoneProd     : return (value * player.pStone)      \ 100
    case ownCrystalProd   : return (value * player.pCrystal)    \ 100
    case ownCreatureProd  : return (value * player.pCreature)   \ 100
    case ownWall          : return (value * player.wall)        \ 100
    case ownTower         : return (value * player.tower)       \ 100
    case oppStone         : return (value * opponent.stone)     \ 100
    case oppCrystal       : return (value * opponent.crystal)   \ 100
    case oppCreature      : return (value * opponent.creature)  \ 100
    case oppStoneProd     : return (value * opponent.pStone)    \ 100
    case oppCrystalProd   : return (value * opponent.pCrystal)  \ 100
    case oppCreatureProd  : return (value * opponent.pCreature) \ 100
    case oppWall          : return (value * opponent.wall)      \ 100
    case oppTower         : return (value * opponent.tower)     \ 100
    case diffStone        : return (value * (player.stone     - opponent.stone))     \ 100
    case diffCrystal      : return (value * (player.crystal   - opponent.crystal))   \ 100
    case diffCreature     : return (value * (player.creature  - opponent.creature))  \ 100
    case diffStoneProd    : return (value * (player.pStone    - opponent.pStone))    \ 100
    case diffCrystalProd  : return (value * (player.pCrystal  - opponent.pCrystal))  \ 100
    case diffCreatureProd : return (value * (player.pCreature - opponent.pCreature)) \ 100
    case diffWall         : return (value * (player.wall      - opponent.wall))      \ 100
    case diffTower        : return (value * (player.tower     - opponent.tower))     \ 100
    case diffOwnBuilding  : return (value * (player.tower     - player.wall))        \ 100
    case diffOppBuilding  : return (value * (opponent.tower   - opponent.wall))      \ 100
  end select
  return value
end function

function getConditionValue(cond as string) as TCond
  select case cond
    case "ownStone"                   : return TCond.ownStone
    case "ownCrystal"                 : return TCond.ownCrystal
    case "ownCreature"                : return TCond.ownCreature
    case "ownStoneProd"               : return TCond.ownStoneProd
    case "ownCrystalProd"             : return TCond.ownCrystalProd
    case "ownCreatureProd"            : return TCond.ownCreatureProd
    case "ownWall"                    : return TCond.ownWall
    case "ownTower"                   : return TCond.ownTower
    case "opponentStone"              : return TCond.oppStone
    case "opponentCrystal"            : return TCond.oppCrystal
    case "opponentCreature"           : return TCond.oppCreature
    case "opponentStoneProd"          : return TCond.oppStoneProd
    case "opponentCrystalProd"        : return TCond.oppCrystalProd
    case "opponentCreatureProd"       : return TCond.oppCreatureProd
    case "opponentWall"               : return TCond.oppWall
    case "opponentTower"              : return TCond.oppTower
    case "differenceStone"            : return TCond.diffStone
    case "differenceCrystal"          : return TCond.diffCrystal
    case "differenceCreature"         : return TCond.diffCreature
    case "differenceStoneProd"        : return TCond.diffStoneProd
    case "differenceCrystalProd"      : return TCond.diffCrystalProd
    case "differenceCreatureProd"     : return TCond.diffCreatureProd
    case "differenceWall"             : return TCond.diffWall
    case "differenceTower"            : return TCond.diffTower
    case "differenceOwnBuilding"      : return TCond.diffOwnBuilding
    case "differenceOpponentBuilding" : return TCond.diffOppBuilding
  end select
  return TCond.none
end function

sub readCards(player as TPlayer, filename as string, language as string)
  ' read cards
  dim as integer file = freefile, id, count = 0
  open exepath & "/cards/" & filename for input as #file
  dim as TypXml xml = xmlParse(, file), xmltemp, xmlIf
  id = val(xml.GetAttr("id"))
  xml = xmlParse(xml.inhalt)
  dim as string target
  dim as TCond cond1, cond2, relation
  dim comp as TComp, compValue as integer
  do while xml.inhalt <> ""
    if xml.tag <> "card" then continue do
    redim preserve player.deck(ubound(player.deck)+1)
    xmltemp = xmlParse(xml.inhalt)
    with player.deck(ubound(player.deck))
      .id = (id-1)*256 + count
      count += 1
      do while xmltemp.tag <> ""
        compValue = 0
        select case xmltemp.tag
          case "name"
            if .cardname = "" or xmltemp.GetAttr("lang") = language then .cardname = xmltemp.inhalt
          case "costStone"
            .cost = valint(xmltemp.inhalt) : .costType = TCond.ownStone
          case "costCrystal"
            .cost = valint(xmltemp.inhalt) : .costType = TCond.ownCrystal
          case "costCreature"
            .cost = valint(xmltemp.inhalt) : .costType = TCond.ownCreature
          case "if"
            cond1 = getConditionValue(xmltemp.GetAttr("condition"))
            cond2 = getConditionValue(xmltemp.GetAttr("compareWith"))
            compValue = valint(xmltemp.GetAttr("compareWith"))
            select case xmltemp.GetAttr("relation")
              case "equal"   : comp = TComp.equal
              case "greater" : comp = TComp.greater
              case "lower"   : comp = TComp.lower
            end select
            xmlIf = xmlParse(xmltemp.inhalt)
            relation = getConditionValue(xmlIf.GetAttr("relation"))
            .addEffect readEffect(player, ubound(player.deck), xmlIf), _
                                  valint(xmlIf.inhalt), cond1, cond2, comp, compValue, relation
            xmlIf = xmlParse(xmlIf.rest)
            do while xmlIf.tag = "description"
              if .effect(.countEffects - 1).desc = "" orelse xmlIf.GetAttr("lang") = language then
                .effect(.countEffects - 1).desc = xmlIf.inhalt
              end if
              xmlIf = xmlParse(xmlIf.rest)
            loop
            if xmlIf.tag = "else" then
              xmlIf = xmlParse(xmlIf.rest)
              relation = getConditionValue(xmlIf.GetAttr("relation"))
              .effect(.countEffects - 1).effectTypeAlt = readEffect(player, ubound(player.deck), xmlIf)
              .effect(.countEffects - 1).powerAlt      = valint(xmlIf.inhalt)
              .effect(.countEffects - 1).relatAlt      = relation
              xmlIf = xmlParse(xmlIf.rest)
              do while xmlIf.tag = "description"
                if .effect(.countEffects - 1).descAlt = "" orelse xmlIf.GetAttr("lang") = language then
                  .effect(.countEffects - 1).descAlt = xmlIf.inhalt
                end if
                xmlIf = xmlParse(xmlIf.rest)
              loop
            end if
        case "again" : .special or= Special.again
        case "againStone"    : .special or= Special.again or Special.stone
        case "againCrystal"  : .special or= Special.again or Special.crystal
        case "againCreature" : .special or= Special.again or Special.creature
        case "description"
          if .countEffects = 0 then
            .desc = xmltemp.inhalt
          elseif .effect(.countEffects - 1).desc = "" orelse xmltemp.GetAttr("lang") = language then
            .effect(.countEffects - 1).desc = xmltemp.inhalt
          end if
        case else
          relation = getConditionValue(xmltemp.GetAttr("relation"))
          .addEffect readEffect(player, ubound(player.deck), xmltemp), valint(xmltemp.inhalt),,,,, relation
        end select
        xmltemp = xmlParse(xmltemp.rest)
      loop
      ' change 'damageXY' to 'changeXY'
      for i as integer = 0 to .countEffects - 1
        with .effect(i)
          if .effectType >= TEffectType.damageOwnWall and .effectType <= TEffectType.damageBothTower then
            .effectType    -= TEffectType.damageOwnWall - TEffectType.changeOwnWall
            .power         *= -1
          end if
          if .effectTypeAlt >= TEffectType.damageOwnWall and .effectTypeAlt <= TEffectType.damageBothTower then
            .effectTypeAlt -= TEffectType.damageOwnWall - TEffectType.changeOwnWall
            .powerAlt      *= -1
          end if
        end with
      next
    end with
    xml = xmlParse(xml.rest)
  loop
  close file
end sub

function readEffect(player as TPlayer, byref c as integer, xml as TypXML) as TEffectType
  with player.deck(c)
    select case xml.tag
      case "addProductionStone"
        select case xml.GetAttr("target")
          case "self"  : return TEffectType.changeOwnStoneProd
          case "enemy" : return TEffectType.changeOppStoneProd
          case "both"  : return TEffectType.changeBothStoneProd
        end select
      case "addProductionCrystal"
        select case xml.GetAttr("target")
          case "self"  : return TEffectType.changeOwnCrystalProd
          case "enemy" : return TEffectType.changeOppCrystalProd
          case "both"  : return TEffectType.changeBothCrystalProd
        end select
      case "addProductionCreature"
        select case xml.GetAttr("target")
          case "self"  : return TEffectType.changeOwnCreatureProd
          case "enemy" : return TEffectType.changeOppCreatureProd
          case "both"  : return TEffectType.changeBothCreatureProd
        end select
      case "addStone"
        select case xml.GetAttr("target")
          case "self"  : return TEffectType.changeOwnStone
          case "enemy" : return TEffectType.changeOppStone
          case "both"  : return TEffectType.changeBothStone
        end select
      case "addCrystal"
        select case xml.GetAttr("target")
          case "self"  : return TEffectType.changeOwnCrystal
          case "enemy" : return TEffectType.changeOppCrystal
          case "both"  : return TEffectType.changeBothCrystal
        end select
      case "addCreature"
        select case xml.GetAttr("target")
          case "self"  : return TEffectType.changeOwnCreature
          case "enemy" : return TEffectType.changeOppCreature
          case "both"  : return TEffectType.changeBothCreature
        end select
      case "stealProductionStone"    : return TEffectType.stealStoneProd
      case "stealProductionCrystal"  : return TEffectType.stealCrystalProd
      case "stealProductionCreature" : return TEffectType.stealCreatureProd
      case "stealStone"              : return TEffectType.stealStone
      case "stealCrystal"            : return TEffectType.stealCrystal
      case "stealCreature"           : return TEffectType.stealCreature
      case "buildWall"
        select case xml.GetAttr("target")
          case "self"  : return TEffectType.changeOwnWall
          case "enemy" : return TEffectType.changeOppWall
          case "both"  : return TEffectType.changeBothWall
        end select
      case "buildTower"
        select case xml.GetAttr("target")
          case "self"  : return TEffectType.changeOwnTower
          case "enemy" : return TEffectType.changeOppTower
          case "both"  : return TEffectType.changeBothTower
        end select
      case "damageWall"
        select case xml.GetAttr("target")
          case "self"  : return TEffectType.damageOwnWall
          case "enemy" : return TEffectType.damageOppWall
          case "both"  : return TEffectType.damageBothWall
        end select
      case "damageTower"
       select case xml.GetAttr("target")
          case "self"  : return TEffectType.damageOwnTower
          case "enemy" : return TEffectType.damageOppTower
          case "both"  : return TEffectType.damageBothTower
        end select
      case "damage"
       select case xml.GetAttr("target")
          case "self"  : return TEffectType.ownDamage
          case "enemy" : return TEffectType.oppDamage
          case "both"  : return TEffectType.bothDamage
        end select
      case else
        print "unknown tag <" & xml.tag & ">"
    end select
  end with
end function

function removeNewline(text as string) as string
  dim ret as string = text, search as integer = instr(ret, "\n")
  do while search
    if mid(ret, search - 1, 1) = "-" then
      ret = left(ret, search - 2) & mid(ret, search + 2)
    else
      ret = left(ret, search - 1) & " " & mid(ret, search + 2)
    end if
    search = instr(search - 1, ret, "\n")
  loop
  return ret
end function

function testCondition(player as TPlayer, opponent as TPlayer, e as TEffect) as integer
  if e.comp = TComp.none then return -1
  dim as TComp comp = e.comp
  dim as integer cond1, cond2
  cond1 = calcRelation(player, opponent, 100, e.cond1)
  cond2 = iif(e.cond2 = TCond.none, e.compValue, calcRelation(player, opponent, 100, e.cond2))
  select case comp
    case TComp.equal   : if cond1 <> cond2 then return 0
    case TComp.greater : if cond1 <= cond2 then return 0
    case TComp.lower   : if cond1 >= cond2 then return 0
  end select
  return -1
end function
