#define VERMAJOR 1
#define VERMINOR 0
#define VERPATCH 150712

#include once "inc/xmlparse.bi"
#include once "crt/string.bi"
#include once "inc/TSNE_V3.bi"
#include once "inc/mages.bi"

const ScoreWall       =     5
const ScoreTower      =    10
const ScoreStone      =     1
const ScoreCrystal    =     1
const ScoreCreature   =     1
const ScoreProduction =    10
const ScoreNewCard    =    10
const ScoreWin        =  1000

type TScore
  as integer move, score
end type

'###############################################################################
' NETWORK
'
dim shared serverID as integer, mutex as any ptr, dataBuffer as string
mutex = mutexcreate

declare sub TSNE_Disconnected(byval V_TSNEID as uinteger)
declare sub TSNE_Connected(byval V_TSNEID as UInteger)
declare sub TSNE_NewData(byval V_TSNEID as uinteger, byref V_Data as string)

type connectionType
  as uinteger G_Client
  declare constructor
  declare constructor(ip as string, port as integer)
  declare destructor
  declare sub shutdown
  declare sub send(msg as string)
  declare sub sendCommand(msg as TCommand, txt as string = "")
end type

constructor connectionType
end constructor
constructor connectionType(ip as string, port as integer)
  dim RV as integer
  RV = TSNE_Create_Client(G_Client, ip, port, @TSNE_Disconnected, @TSNE_Connected, @TSNE_NewData) ' create Client
  if RV = TSNE_Const_NoError then
    print "searching for server"
  else
    print "ERROR: " & RV
  end if
end constructor

destructor connectionType
  shutdown
end destructor

sub connectionType.shutdown
  TSNE_Disconnect(G_Client)
  TSNE_WaitClose(G_Client)
end sub

sub connectionType.send(msg as string)
  mutexlock mutex
  TSNE_Data_Send(serverID, msg)
  mutexunlock mutex
end sub

sub connectionType.sendCommand(msg as TCommand, txt as string = "")
  this.send chr(msg) & txt
end sub

'###############################################################################

declare function calcScore(player as TPlayer, opponent as TPlayer) as integer
declare function makeTurn(player as TPlayer, opponent as TPlayer, spec as integer) as TScore
declare function playable(player as TPlayer, cardnumber as integer, spec as integer) as integer
declare function playCard(player as TPlayer, opponent as TPlayer, cardnumber as integer, spec as integer) as integer

declare function readMessageData as string

dim shared as connectionType myClient = connectionType("127.0.0.1", 6006)
dim as string message, tempString
dim as integer countCards, id
dim as TPlayer player, opponent
dim as string filename = dir(exepath & "/cards/*.xml")
do while len(filename)
  readCards(player, filename, "")
  filename = dir
loop
print "waiting for server response ..."

dim as integer spec
dim as TScore cMove

'##########################################################################
' START THE GAME
do
  message &= readMessageData
  if len(message) then
    select case message[0]
      case TCommand.connectionLost
        print "ERROR: Connection lost"
        exit do
' Client gets current game settings
      case TCommand.submitSettings
        if len(message) < 2 then
          myClient.sendCommand TCommand.fault
          message = ""
          continue do
        end if
        countCards = message[1]
        if countCards > MAXCARDS or len(message) < countCards*2 + 37 then
          myClient.sendCommand TCommand.fault
          message = ""
          continue do
        end if
        for i as integer = 0 to countCards-1
          id = (message[i*2+2] shl 8) + message[i*2+3]
          for k as integer = 0 to ubound(player.deck)
            if player.deck(k).id = id then
              player.card(i) = k
              exit for
            end if
          next
        next
        player.stone     = (message[countCards*2+2] shl 7) + message[countCards*2+3]
        player.pStone    = message[countCards*2+4]
        player.crystal   = (message[countCards*2+5] shl 7) + message[countCards*2+6]
        player.pCrystal  = message[countCards*2+7]
        player.creature  = (message[countCards*2+8] shl 7) + message[countCards*2+9]
        player.pCreature = message[countCards*2+10]
        player.tower     = (message[countCards*2+11] shl 7) + message[countCards*2+12]
        player.wall      = (message[countCards*2+13] shl 7) + message[countCards*2+14]
        opponent.stone     = (message[countCards*2+15] shl 7) + message[countCards*2+16]
        opponent.pStone    = message[countCards*2+17]
        opponent.crystal   = (message[countCards*2+18] shl 7) + message[countCards*2+19]
        opponent.pCrystal  = message[countCards*2+20]
        opponent.creature  = (message[countCards*2+21] shl 7) + message[countCards*2+22]
        opponent.pCreature = message[countCards*2+23]
        opponent.tower     = (message[countCards*2+24] shl 7) + message[countCards*2+25]
        opponent.wall      = (message[countCards*2+26] shl 7) + message[countCards*2+27]
        setting.maxStone    = (message[countCards*2+28] shl 7) + message[countCards*2+29]
        setting.maxCrystal  = (message[countCards*2+30] shl 7) + message[countCards*2+31]
        setting.maxCreature = (message[countCards*2+32] shl 7) + message[countCards*2+33]
        setting.maxTower    = (message[countCards*2+34] shl 7) + message[countCards*2+35]
        spec                = message[countCards*2+36]
        message = mid(message, countCards*2+38)
' client gets command to play
      case TCommand.yourTurn
        message = mid(message, 2)
        cmove = makeTurn(player, opponent, spec and (Special.stone or Special.crystal or Special.creature))
        if cmove.move < 0 then
          myClient.sendCommand TCommand.dropCard, chr(-cmove.move-1)
        else
          myClient.sendCommand TCommand.playCard, chr(cmove.move)
        end if
' cases where's nothing interesting to do
      case TCommand.wait4Turn
        message = mid(message, 2)
      case TCommand.takeCard
        message = mid(message, 5)
      case TCommand.playCard, TCommand.dropCard
        message = mid(message, 4)
      case TCommand.gameWin, TCommand.gameLose, TCommand.gameDraw
        exit do
' unknown command
      case else
        dim as integer i
        for i = 1 to len(message)-1
          if message[i] > 127 then exit for
        next
        message = mid(message, i+1)
    end select
  end if
loop until inkey = chr(255, 107)

myClient.shutdown
'threadwait serverthread
mutexdestroy mutex
end

sub TSNE_Disconnected(byval V_TSNEID as uinteger)
  mutexlock mutex
  dataBuffer = chr(TCommand.connectionLost)
  mutexunlock mutex
end sub

sub TSNE_Connected(byval V_TSNEID as uinteger)
  mutexlock mutex
  serverID = V_TSNEID
  mutexunlock mutex
end sub

sub TSNE_NewData(byval V_TSNEID as uinteger, byref V_Data as string)
  mutexlock mutex
  dataBuffer &= V_Data
  mutexunlock mutex
end sub

function calcScore(player as TPlayer, opponent as TPlayer) as integer
  dim as integer ret
  ret = (player.wall - opponent.wall) * ScoreWall
  ret += (player.tower - opponent.tower) * ScoreTower
  ret += (player.stone - opponent.stone + (player.pStone - opponent.pStone)*ScoreProduction)*ScoreStone
  ret += (player.crystal - opponent.crystal + (player.pCrystal - opponent.pCrystal)*ScoreProduction)*ScoreCrystal
  ret += (player.creature - opponent.creature + (player.pCreature - opponent.pCreature)*ScoreProduction)*ScoreCreature
  if player.tower >= setting.maxTower or opponent.tower <= 0 or player.stone >= setting.maxStone _
     or player.crystal >= setting.maxCrystal or player.creature >= setting.maxCreature then ret += ScoreWin
  if opponent.tower >= setting.maxTower or player.tower <= 0 or opponent.stone >= setting.maxStone _
     or opponent.crystal >= setting.maxCrystal or opponent.creature >= setting.maxCreature then ret -= ScoreWin
  return ret
end function

function makeTurn(player as TPlayer, opponent as TPlayer, spec as integer) as TScore
  dim as TPlayer pNew, oNew
  dim as integer maxScore = calcScore(pNew, oNew), score, bestMove = 0
  for i as integer = 0 to 5
    if player.card(i) > 0 then
      pNew = player
      oNew = opponent
      if playCard(pNew, oNew, i, spec) then
        if pNew.deck(pNew.card(i)).special and Special.again then
          pNew.card(i) = 0
          score = makeTurn(pNew, oNew, pNew.deck(pNew.card(i)).special and _
                  (Special.stone or Special.crystal or Special.creature)).score + ScoreNewCard
        else
          score = calcScore(pNew, oNew)
        end if
        if score > maxScore then
          maxScore = score
          bestMove = i
        end if
      end if
    end if
  next
  if bestMove then
    playCard(pNew, oNew, bestMove, spec)
  else
    ' TODO: discard
    bestmove = -int(rnd*6) - 1
    score = calcScore(pNew, oNew)
  end if
  return type<TScore>(bestMove, score)
end function

function playable(p as TPlayer, cardnumber as integer, spec as integer) as integer
  dim as TCard c = p.deck(p.card(cardnumber))
  if spec = (Special.again or Special.stone)    andalso c.costType <> TCond.ownStone    then return 0
  if spec = (Special.again or Special.crystal)  andalso c.costType <> TCond.ownCrystal  then return 0
  if spec = (Special.again or Special.creature) andalso c.costType <> TCond.ownCreature then return 0
  select case c.costType
    case TCond.ownStone    : if c.cost > p.stone    then return 0
    case TCond.ownCrystal  : if c.cost > p.crystal  then return 0
    case TCond.ownCreature : if c.cost > p.creature then return 0
  end select
  return -1
end function

function playCard(player as TPlayer, opponent as TPlayer, cardnumber as integer, spec as integer) as integer
  dim as TCard c = player.deck(player.card(cardnumber))
  if playable(player, cardnumber, spec) = 0 then return 0
  select case c.costType
    case TCond.ownStone    : player.stone    -= c.cost
    case TCond.ownCrystal  : player.crystal  -= c.cost
    case TCond.ownCreature : player.creature -= c.cost
  end select
  for i as integer = 0 to c.countEffects - 1
    calcEffect(player, opponent, c.effect(i))
  next
  with player
    if .stone     < 0 then .stone     = 0
    if .crystal   < 0 then .crystal   = 0
    if .creature  < 0 then .creature  = 0
    if .pStone    < 1 then .pStone    = 1
    if .pCrystal  < 1 then .pCrystal  = 1
    if .pCreature < 1 then .pCreature = 1
    if .wall      < 0 then .wall      = 0
    if .tower     < 0 then .tower     = 0
  end with
  with opponent
    if .stone     < 0 then .stone     = 0
    if .crystal   < 0 then .crystal   = 0
    if .creature  < 0 then .creature  = 0
    if .pStone    < 1 then .pStone    = 1
    if .pCrystal  < 1 then .pCrystal  = 1
    if .pCreature < 1 then .pCreature = 1
    if .wall      < 0 then .wall      = 0
    if .tower     < 0 then .tower     = 0
  end with
  return -1
end function

function readMessageData as string
  dim ret as string, length as integer
  mutexlock mutex
  ret = dataBuffer
  dataBuffer = ""
  mutexunlock mutex
  return ret
end function
