#define VERMAJOR  1
#define VERMINOR  0
#define VERPATCH  150712
const VERSION_STRING = VERMAJOR & "." & VERMINOR & "." & VERPATCH

#include once "inc/xmlparse.bi"
#include once "inc/mages.bi"
#include once "crt/string.bi"
#include once "inc/TSNE_V3.bi"

declare function readMessageData(i as integer) as string
declare function playable(p as TPlayer, cardnumber as integer, spec as integer) as integer
declare function playCard(player as TPlayer, opponent as TPlayer, cardnumber as integer, spec as integer) as integer
declare function settingString(player as TPlayer, opponent as TPlayer, spec as integer) as string
declare function testGameOver(player() as TPlayer) as integer

randomize
dim file as integer, tempstring as string

'###############################################################################
' CREATE SETTINGS
' TODO: make it possible to change by server commands
with setting
  ' default settings
  .startPStone    = 1
  .startPCrystal  = 1
  .startPCreature = 1
  .startStone     = 5
  .startCrystal   = 5
  .startCreature  = 5
  .maxStone       = 500
  .maxCrystal     = 500
  .maxCreature    = 500
  .startTower     = 20
  .startWall      = 10
  .maxTower       = 100
  ' read settings from file
  file = freefile
  open exepath & "/setting.txt" for input as file
  do until eof(file)
    line input #file, tempstring
    tempstring = trim(lcase(tempstring))
    if left(tempstring, 20) = "startproductionstone"    then .startPStone    = valint(mid(tempstring, instr(tempstring, "=")+1))
    if left(tempstring, 22) = "startproductioncrystal"  then .startPCrystal  = valint(mid(tempstring, instr(tempstring, "=")+1))
    if left(tempstring, 23) = "startproductioncreature" then .startPCreature = valint(mid(tempstring, instr(tempstring, "=")+1))
    if left(tempstring, 19) = "startresourcesstone"     then .startStone     = valint(mid(tempstring, instr(tempstring, "=")+1))
    if left(tempstring, 21) = "startresourcescrystal"   then .startCrystal   = valint(mid(tempstring, instr(tempstring, "=")+1))
    if left(tempstring, 22) = "startresourcescreature"  then .startCreature  = valint(mid(tempstring, instr(tempstring, "=")+1))
    if left(tempstring, 12) = "maximalstone"    then .maxStone    = valint(mid(tempstring, instr(tempstring, "=")+1))
    if left(tempstring, 14) = "maximalcrystal"  then .maxCrystal  = valint(mid(tempstring, instr(tempstring, "=")+1))
    if left(tempstring, 15) = "maximalcreature" then .maxCreature = valint(mid(tempstring, instr(tempstring, "=")+1))
    if left(tempstring, 10) = "starttower"      then .startTower  = valint(mid(tempstring, instr(tempstring, "=")+1))
    if left(tempstring,  9) = "startwall"       then .startWall   = valint(mid(tempstring, instr(tempstring, "=")+1))
    if left(tempstring,  8) = "maxtower"        then .maxTower    = valint(mid(tempstring, instr(tempstring, "=")+1))
  loop
  close #file
end with
dim shared as TPlayer player(1)
readCards(player(0), "default.xml", "")
readCards(player(0), "construction.xml", "")
readCards(player(0), "magic.xml", "")
readCards(player(1), "default.xml", "")
readCards(player(1), "construction.xml", "")
readCards(player(1), "magic.xml", "")
dim as TDeck ptr deckPosition(1)
deckPosition(0) = new TDeck
deckPosition(1) = new TDeck

print "SERVER FOR MAGES v" & VERSION_STRING
print "Print ESC to shutdown the server."
print
'###############################################################################
' CREATE SERVER
dim shared countConnections as integer, mutex as any ptr, dataBuffer(1) as string, id(1) as uinteger

sub TSNE_Disconnected(byval V_TSNEID as uinteger)
  print "* " & V_TSNEID & " disconnected"
  print "* server resetted"
  print string(75, "*")
  mutexlock mutex
  for i as integer = 0 to 1
    TSNE_Data_Send id(i), chr(TCommand.connectionLost)    
  next
  countConnections = 0
  dataBuffer(0) = chr(TCommand.restart)
  mutexunlock mutex
end sub

sub TSNE_Connected(byval V_TSNEID as uinteger)
  print "* trying to connect: ID " & V_TSNEID
  mutexlock mutex
  dim as integer count = countConnections
  mutexunlock mutex
  if count > 1 then
    print "* rejected (to many connections)"
    TSNE_Data_Send V_TSNEID, chr(TCommand.tooManyClients)
  end if
  mutexlock mutex
  id(count) = V_TSNEID
  countConnections += 1
  mutexunlock mutex
  if count = 0 then
    TSNE_Data_Send V_TSNEID, chr(TCommand.wait4connection)
  else
    TSNE_Data_Send V_TSNEID, chr(TCommand.ready)
    print "* a new game will start now"
    print string(75, "*")
    mutexlock mutex
    TSNE_Data_Send id(0), chr(TCommand.ready)
    dataBuffer(0) = chr(TCommand.ready)
    mutexunlock mutex
  end if
end sub

sub TSNE_NewConnectionCanceled(byval V_TSNEID as uinteger, byval V_IPA as string)
  print "* " & V_TSNEID & ": connection canceled"
end sub

sub TSNE_NewData(byval V_TSNEID as uinteger, byref V_Data as string)
  mutexlock mutex
  select case V_TSNEID
    case id(0) : dataBuffer(0) &= V_Data
    case id(1) : dataBuffer(1) &= V_Data
  end select
  mutexunlock mutex
end sub

sub TSNE_NewConnection(byval V_TSNEID as uinteger, byval V_RequestID as Socket, byval V_IPA as string)
  print "* client tries to connect"
  dim XNewTSNEID as uinteger
  dim XReturnIPA as string
  TSNE_Create_Accept(V_RequestID, XNewTSNEID, XReturnIPA, @TSNE_Disconnected, @TSNE_Connected, @TSNE_NewData)
end sub

type connectionType
  as uinteger G_Server
  declare constructor(port as ushort)
  declare destructor
  declare sub host
  declare sub shutdown
  declare sub send(id as uinteger, message as string)
  declare sub sendCommand(number as integer, msg as TCommand, txt as string = "")
end type

constructor connectionType(port as ushort)
  dim RV as integer
  RV = TSNE_Create_Server(G_Server, port, 1, @TSNE_NewConnection, @TSNE_NewConnectionCanceled)
  if RV = TSNE_Const_NoError then
    print "* server ready"
  else
    print "* ERROR: " & RV
  end if
end constructor

destructor connectionType
  shutdown
end destructor

sub connectionType.shutdown
  sendCommand 0, TCommand.connectionLost
  sendCommand 1, TCommand.connectionLost
  print "shutting down, please wait ..."
  sleep 2000
  TSNE_Disconnect(G_Server)
  TSNE_WaitClose(G_Server)
  print "... done"
end sub

sub connectionType.send(id as uinteger, message as string)
  TSNE_Data_Send(id, message)
end sub

sub connectionType.sendCommand(number as integer, msg as TCommand, txt as string = "")
  mutexlock mutex
  this.send(id(number), chr(msg) & txt)
  mutexunlock mutex
end sub

dim as connectionType myServer = 6006

'###############################################################################
' MAIN
dim as string  message(1), curSettings, msg
dim as integer position, spec
dim as TCard   tempCard
do
  ' server main loop - waiting for connections
  sleep 1
  message(0) = readMessageData(0)
  message(1) = readMessageData(1)
  if message(0) = chr(TCommand.ready) then
    dim as integer currentPlayer = int(rnd*2), spec
    ' create card decks and player settings
    for p as integer = 0 to 1
      with player(p)
        .pStone    = setting.startPStone
        .pCrystal  = setting.startPCrystal
        .pCreature = setting.startPCreature
        .stone     = setting.startStone
        .crystal   = setting.startCrystal
        .creature  = setting.startCreature
        .tower     = setting.startTower
        .wall      = setting.startWall
      end with
      for i as integer = 0 to ubound(player(p).deck) - 1
        for k as integer = i+1 to ubound(player(p).deck) - 1
          if rnd*(k-i+1) < 1 then swap player(p).deck(i), player(p).deck(k)
        next
        deckPosition(p)->add(i)
      next
      dim as TDeck ptr start = deckPosition(p)->nxt
      for i as integer = 0 to 5
        player(p).card(i) = deckPosition(p)->nxt->card
        deckPosition(p)->del
      next
    next

    do
      ' game main loop

      ' sending current game data
      myServer.sendCommand currentPlayer, TCommand.submitSettings, _
                           settingString(player(currentPlayer), player(1-currentPlayer), spec)
      myServer.sendCommand 1-currentPlayer, TCommand.submitSettings, _
                           settingString(player(1-currentPlayer), player(currentPlayer), spec)
      ' check win situation
      select case testGameOver(player())
        case 1
          myServer.sendCommand 0, TCommand.gameWin
          myServer.sendCommand 1, TCommand.gameLose
        case 2
          myServer.sendCommand 0, TCommand.gameLose
          myServer.sendCommand 1, TCommand.gameWin
        case 3
          myServer.sendCommand 0, TCommand.gameDraw
          myServer.sendCommand 1, TCommand.gameDraw
        case else
          myServer.sendCommand currentPlayer, TCommand.wait4turn
          currentPlayer = 1-currentPlayer
          myServer.sendCommand currentPlayer, TCommand.yourTurn
      end select
      ' waiting for response
      do
        sleep 1
        if inkey = chr(27) then exit do, do, do
        for p as integer = 0 to 1
          message(p) &= readMessageData(p)
          if len(message(p)) then
            select case message(p)[0]
              case TCommand.withdraw
                exit do, do
              case TCommand.restart
                exit do, do
              case TCommand.submitSettings
                dim as string msg = settingString(player(p), player(1-p), spec)
                if p = currentPlayer then
                  myServer.sendCommand currentPlayer, TCommand.submitSettings, msg
                else
                  myServer.sendCommand currentPlayer, TCommand.wait4turn, msg
                end if
              case TCommand.playCard, dropCard
                if p = currentPlayer then
                  if len(message(p)) < 2 then
                    myServer.sendCommand currentPlayer, TCommand.fault, chr(TError.missingParameter)
                    message(p) = ""
                    continue for
                  end if
                  if message(p)[1] > MAXCARDS-1 then
                    myServer.sendCommand currentPlayer, TCommand.fault, chr(TError.wrongParameter)
                    message(p) = mid(message(p), 3)
                    continue for
                  end if
                  position = message(p)[1]
                  tempCard = player(p).deck(player(p).card(position))
                  if message(p)[0] = TCommand.playCard then
                    if playable(player(p), position, spec) = 0 then
                      myServer.sendCommand p, TCommand.fault, chr(TError. cardNotPlayable)
                      message(p) = mid(message(p), 3)
                      continue for
                    end if
                    playCard player(currentPlayer), player(1-currentPlayer), position, spec
                  end if
                  with player(p)
                    myServer.sendCommand 1-p, message(p)[0], _
                             chr(.deck(.card(position)).id shr 8, .deck(.card(position)).id)
                    deckPosition(p)->add(.card(position))
                    deckPosition(p) = deckPosition(p)->nxt
                    .card(position) = deckPosition(p)->nxt->card
                    myServer.sendCommand currentPlayer, TCommand.takeCard, chr(position, _
                                         .deck(.card(position)).id shr 8, .deck(.card(position)).id and 255)
                    deckPosition(p)->del
                  end with
                  if tempCard.special and (Special.again or Special.stone or Special.crystal or Special.creature) then
                    spec = tempcard.special
                    currentPlayer = 1-p
                  else
                    spec = Special.nothing
                    with player(1-p)
                      .stone    += .pStone
                      .crystal  += .pCrystal
                      .creature += .pCreature
                    end with
                  end if
                  message(p) = mid(message(p), 3)
                  exit do
                else
                  myServer.sendCommand currentPlayer, TCommand.fault, chr(TError.notYourTurn)
                  message(p) = mid(message(p), 3)
                end if
              case else
                dim as integer i
                for i = 1 to len(message(p))-1
                  if message(p)[i] > 127 then exit for
                next
                message(p) = mid(message(p), i+1)
            end select
          end if
        next
      loop
    loop

  end if
loop until inkey = chr(27)

' destruct
myServer.shutdown
mutexdestroy mutex
for p as integer = 0 to 1
  do while deckPosition(p)->nxt > 0
    deckPosition(p)->del
  loop
next
end

function readMessageData(i as integer) as string
  dim ret as string, length as integer
  mutexlock mutex
  ret = dataBuffer(i)
  dataBuffer(i) = ""
  mutexunlock mutex
  return ret
end function

function playable(p as TPlayer, cardnumber as integer, spec as integer) as integer
  dim as TCard c = p.deck(p.card(cardnumber))
  if spec = Special.stone    andalso c.costType <> TCond.ownStone    then return 0
  if spec = Special.crystal  andalso c.costType <> TCond.ownCrystal  then return 0
  if spec = Special.creature andalso c.costType <> TCond.ownCreature then return 0
  select case c.costType
    case TCond.ownStone    : if c.cost > p.stone    then return 0
    case TCond.ownCrystal  : if c.cost > p.crystal  then return 0
    case TCond.ownCreature : if c.cost > p.creature then return 0
  end select
  return -1
end function

function playCard(player as TPlayer, opponent as TPlayer, cardnumber as integer, spec as integer) as integer
  dim as TCard c = player.deck(player.card(cardnumber))
  if playable(player, cardnumber, spec) = 0 then return 0
  select case c.costType
    case TCond.ownStone    : player.stone    -= c.cost
    case TCond.ownCrystal  : player.crystal  -= c.cost
    case TCond.ownCreature : player.creature -= c.cost
  end select
  for i as integer = 0 to c.countEffects - 1
    calcEffect(player, opponent, c.effect(i))
  next
  with player
    if .stone     < 0 then .stone     = 0
    if .crystal   < 0 then .crystal   = 0
    if .creature  < 0 then .creature  = 0
    if .pStone    < 1 then .pStone    = 1
    if .pCrystal  < 1 then .pCrystal  = 1
    if .pCreature < 1 then .pCreature = 1
    if .wall      < 0 then .wall      = 0
    if .tower     < 0 then .tower     = 0
  end with
  with opponent
    if .stone     < 0 then .stone     = 0
    if .crystal   < 0 then .crystal   = 0
    if .creature  < 0 then .creature  = 0
    if .pStone    < 1 then .pStone    = 1
    if .pCrystal  < 1 then .pCrystal  = 1
    if .pCreature < 1 then .pCreature = 1
    if .wall      < 0 then .wall      = 0
    if .tower     < 0 then .tower     = 0
  end with
  return -1
end function

function settingString(player as TPlayer, opponent as TPlayer, spec as integer) as string
  dim as string ret
  ret = chr(MAXCARDS)
  for i as integer = 0 to MAXCARDS-1
    dim as integer id = player.deck(player.card(i)).id
    ret &= chr(id shr 8, id and 255)
  next
  with player
    ret &= chr(.stone    shr 7, .stone    and 127, .pStone)
    ret &= chr(.crystal  shr 7, .crystal  and 127, .pCrystal)
    ret &= chr(.creature shr 7, .creature and 127, .pCreature)
    ret &= chr(.tower shr 7, .tower and 127, .wall shr 7, .wall and 127)
  end with
  with opponent
    ret &= chr(.stone    shr 7, .stone    and 127, .pStone)
    ret &= chr(.crystal  shr 7, .crystal  and 127, .pCrystal)
    ret &= chr(.creature shr 7, .creature and 127, .pCreature)
    ret &= chr(.tower shr 7, .tower and 127, .wall shr 7, .wall and 127)
  end with
  ret &= chr(setting.maxStone    shr 7, setting.maxStone    and 127)
  ret &= chr(setting.maxCrystal  shr 7, setting.maxCrystal  and 127)
  ret &= chr(setting.maxCreature shr 7, setting.maxCreature and 127)
  ret &= chr(setting.maxTower    shr 7, setting.maxTower    and 127)
  ret &= chr(spec)
  return ret
end function

function testGameOver(player() as TPlayer) as integer
  dim as integer ret = -(player(1).tower <= 0) - 2*(player(0).tower <= 0)
  if ret = 0 then
    if player(1).tower <= 0 or player(0).tower >= setting.maxTower _
       or player(0).stone >= setting.maxStone _
       or player(0).crystal >= setting.maxCrystal _
       or player(0).creature >= setting.maxCreature _
       then ret = 1
    if player(0).tower <= 0 or player(1).tower >= setting.maxTower _
       or player(1).stone >= setting.maxStone _
       or player(1).crystal >= setting.maxCrystal _
       or player(1).creature >= setting.maxCreature _
       then ret += 2
  end if
  return ret
end function
